import pkgutil
from dataclasses import dataclass
from importlib.machinery import FileFinder
from os.path import join
from typing import Iterable, Optional, Set


@dataclass(frozen=True)
class ModuleData:
    module_name: str
    module_path: str


def discover_modules(base_path: str, exclude_modules: Optional[Set[str]] = None) -> Iterable[ModuleData]:
    if exclude_modules is None:
        exclude_modules = set()

    for module_finder, module_name, is_pkg in pkgutil.iter_modules([base_path]):
        if module_name in exclude_modules:
            continue

        if not isinstance(module_finder, FileFinder):
            continue

        module_path = join(module_finder.path, module_name)
        if is_pkg:
            yield from discover_modules(module_path, exclude_modules)
        else:
            yield ModuleData(module_name=module_name, module_path=module_path)



from typing import Type


class PrimaryDependencyAlreadyDefinedError(Exception):
    def __init__(self, base_class: Type):
        self.base_class = base_class
        super().__init__(f"Primary dependency for base {base_class.__name__} already defined")



# pytype: disable=invalid-annotation
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Type, TypeVar

from yandil.argument import Argument

DT = TypeVar("DT")


@dataclass
class Dependency:
    cls: Type[DT]
    arguments: List[Argument] = field(default_factory=list, init=False)
    value: Optional[DT] = field(default=None)
    is_resolved: bool = field(default=False)
    is_primary: Optional[bool] = field(default=None)

    def resolve(self) -> Type:
        if not self.is_resolved:
            args = self.__resolve_args()
            self.value = self.cls(**args)
            self.is_resolved = True

        return self.value

    def __resolve_args(self) -> Dict[str, Any]:
        return {argument.name: argument.value for argument in self.arguments}



from dataclasses import dataclass, field
from os import environ
from typing import Optional


@dataclass
class Environment:
    variable_name: str
    value: Optional[str] = field(default=None)
    is_resolved: bool = field(default=False)

    def resolve(self) -> str:
        if self.is_resolved is False:
            self.value = self.__get_value_from_env()

        self.is_resolved = True
        return self.value

    def __get_value_from_env(self) -> str:
        return environ[self.variable_name]



# pytype: disable=attribute-error
from typing import Callable, List, Set, Tuple, Type, TypeVar, get_origin

__SUPPORTED_ITERABLE_TYPE_HINTS = {get_origin(List), get_origin(Set), get_origin(Tuple)}


def is_type_hint_iterable(type_hint: Type) -> bool:
    if (origin := get_origin(type_hint)) is not None:
        return origin in __SUPPORTED_ITERABLE_TYPE_HINTS
    return False


def type_hint_iterable_builder_factory(iterator_type_hint: Type) -> Callable:
    match iterator_type_hint.__name__:
        case List.__name__:
            return list
        case Set.__name__:
            return set
        case Tuple.__name__:
            return tuple
        case _:
            raise NotImplementedError(f"Unsupported iterator type hint {iterator_type_hint.__name__}")


TT = TypeVar("TT")


def str_to_builtin_type(value: str, target_type: Type[TT]) -> TT:
    if target_type is bool:
        return str_to_bool(value)
    return target_type(value)


def str_to_bool(value: str) -> bool:
    value = value.lower()
    if value in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif value in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError(f"Invalid truth value {value}")



from abc import ABC
from collections import defaultdict
from dataclasses import is_dataclass
from inspect import Parameter, signature
from typing import (
    Any,
    Dict,
    Final,
    Generic,
    Iterable,
    List,
    Optional,
    Protocol,
    Set,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from yandil.argument import Argument
from yandil.configuration.configuration_container import ConfigurationContainer, default_configuration_container
from yandil.dependency import Dependency
from yandil.errors.abstract_class_not_allowed_error import AbstractClassNotAllowedError
from yandil.errors.configuration_value_type_mismatch_error import ConfigurationValueTypeMismatchError
from yandil.errors.dependency_not_found_error import DependencyNotFoundError
from yandil.errors.missing_configuration_value_error import MissingConfigurationValueError
from yandil.errors.missing_type_hint_item_type_error import MissingTypeHintItemTypeError
from yandil.errors.primary_dependency_already_defined_error import PrimaryDependencyAlreadyDefinedError
from yandil.errors.primary_dependency_not_found_error import PrimaryDependencyNotFoundError
from yandil.typing_tools import is_type_hint_iterable, str_to_builtin_type, type_hint_iterable_builder_factory

DT = TypeVar("DT")


class Container:
    __EXCLUDED_BASES: Final[Set[str]] = {object}
    __ABSTRACT_BASES: Final[Set[str]] = {Protocol, Generic, ABC}
    __BUILTIN_TYPES: Final[Set[str]] = {int, float, str, bool, bytes, bytearray}

    def __init__(self, configuration_container: ConfigurationContainer):
        self.__dependency_map: Dict[Type, Dependency] = {}
        self.__bases_map: Dict[Type, List[Type]] = defaultdict(list)
        self.__configuration_container = configuration_container

    def add(self, cls: Type, is_primary: Optional[bool] = None) -> None:
        if self.__is_abstract_class(cls):
            raise AbstractClassNotAllowedError(cls)

        dependency = Dependency(cls, is_primary=is_primary)

        self.__update_parents_map(cls, is_primary)
        self.__dependency_map[cls] = dependency

    def __setitem__(self, cls: Type[DT], value: DT) -> None:
        if self.__is_abstract_class(cls):
            cls = value.__class__
            is_primary = True
            dependency = Dependency(cls, value=value, is_resolved=True, is_primary=True)
        else:
            is_primary = None
            dependency = Dependency(cls, value=value, is_resolved=True)

        self.__update_parents_map(cls, is_primary)
        self.__dependency_map[cls] = dependency

    def __is_abstract_class(self, cls: Type) -> bool:
        for base in cls.__bases__:
            if base in self.__ABSTRACT_BASES:
                return True
        return False

    def __update_parents_map(self, cls: Type, is_primary: Optional[bool]) -> None:
        for base in cls.__bases__:
            if base in self.__EXCLUDED_BASES:
                continue

            if (
                is_primary is True
                and base in self.__bases_map
                and self.__get_primary_dependency_from_base_children(self.__bases_map[base])
            ):
                raise PrimaryDependencyAlreadyDefinedError(base)

            self.__bases_map[base].append(cls)

    def __getitem__(self, cls: Type[DT]) -> Optional[DT]:
        dependency = self.__dependency_map.get(cls, None)
        if dependency is None and cls in self.__bases_map:
            dependency = self.__get_dependency_from_base(cls)

        if dependency is not None:
            self.__set_arguments(dependency)
            return dependency.resolve()

        if dependency is None and self.__is_optional(cls):
            return self.__get_optional_dependency(cls)

        raise DependencyNotFoundError(cls)

    def __get_primary_dependency_from_base_children(self, base_children: List[Type]) -> Optional[Dependency]:
        for base_child in base_children:
            dependency = self.__dependency_map[base_child]
            if dependency is not None and dependency.is_primary:
                return dependency

        return None

    def __get_dependency_from_base(self, cls: Type) -> Dependency:
        base_children = self.__bases_map[cls]
        if len(base_children) == 1:
            return self.__dependency_map[base_children[0]]

        primary_dependency = self.__get_primary_dependency_from_base_children(base_children)
        if primary_dependency is None:
            raise PrimaryDependencyNotFoundError(cls)
        return primary_dependency

    def __is_optional(self, cls: Type) -> bool:
        if get_origin(cls) is not Union:
            return False

        typing_args = get_args(cls)
        if not typing_args:
            raise MissingTypeHintItemTypeError()

        return len(typing_args) == 2 and type(None) in typing_args

    def __get_optional_dependency(self, cls: Type[DT]) -> Optional[DT]:
        optional_inner_type = next(arg for arg in get_args(cls))
        try:
            return self[optional_inner_type]
        except DependencyNotFoundError:
            return None

    def __set_arguments(self, dependency: Dependency) -> None:
        if len(dependency.arguments) > 0:
            return

        for argument_name, argument_type in get_type_hints(dependency.cls.__init__).items():
            if self.__should_skip_argument(dependency.cls, argument_name):
                continue

            if self.__is_configuration_argument(argument_type):
                configuration_value = self.__configuration_container[argument_name]
                if configuration_value is None:
                    raise MissingConfigurationValueError(argument_name)

                dependency.arguments.append(
                    Argument(
                        argument_name,
                        self.__adapt_configuration_value_type(argument_type, configuration_value, argument_name),
                    )
                )
                continue

            if is_type_hint_iterable(argument_type):
                argument_value = self.__get_iterable_dependency_value(argument_type)
                dependency.arguments.append(Argument(argument_name, value=argument_value))
                continue

            try:
                argument_value = self[argument_type]
            except DependencyNotFoundError as err:
                dependency_default_value = self.__get_keyword_dependency_default_value(dependency.cls, argument_name)
                if dependency_default_value is None:
                    raise err
                dependency.arguments.append(Argument(argument_name, value=dependency_default_value))
                continue

            dependency.arguments.append(Argument(argument_name, value=argument_value))

    def __is_configuration_argument(self, argument_type: str) -> bool:
        return argument_type in self.__BUILTIN_TYPES

    def __adapt_configuration_value_type(
        self, argument_type: type, configuration_value: Any, configuration_value_name
    ) -> Any:
        actual_type = type(configuration_value)
        if argument_type == actual_type:
            return configuration_value
        if actual_type != str:
            raise ConfigurationValueTypeMismatchError(configuration_value_name, argument_type, actual_type)
        return str_to_builtin_type(configuration_value, argument_type)

    def __should_skip_argument(self, dependency_type: Type, argument_name: str) -> bool:
        if is_dataclass(dependency_type):
            return argument_name == "return"
        return argument_name == "self"

    def __get_iterable_dependency_value(self, iterable_type_hint: Type) -> Iterable:
        type_hint_args = get_args(iterable_type_hint)
        if not type_hint_args:
            raise MissingTypeHintItemTypeError()

        inner_iterable_type = get_args(iterable_type_hint)[0]
        dependency_types = self.__bases_map[inner_iterable_type]
        return type_hint_iterable_builder_factory(iterable_type_hint)(
            self[dependency_type] for dependency_type in dependency_types
        )

    def __get_keyword_dependency_default_value(self, cls: Type, dependency_name: str) -> Optional[Any]:
        init_signature = signature(cls.__init__)
        dependency_param = init_signature.parameters.get(dependency_name)
        if dependency_param is None or dependency_param.default is Parameter.empty:
            return None
        return dependency_param.default


default_container = Container(configuration_container=default_configuration_container)



from ast import Call, ClassDef, FunctionDef, Name, Subscript, parse
from dataclasses import dataclass
from importlib import import_module
from os import path
from os.path import relpath, splitext
from typing import Iterable, Set, Type


@dataclass(frozen=True)
class ClassData:
    module_file_path: str
    class_name: str

    def to_class(self, sources_root_path: str) -> Type:
        module_path_without_ext = splitext(self.module_file_path)[0]
        module_rel_path = relpath(module_path_without_ext, sources_root_path)
        module = import_module(module_rel_path.replace(path.sep, "."))
        return getattr(module, self.class_name)


def discover_classes_from_module(module_file_path: str) -> Iterable[ClassDef]:
    with open(module_file_path, "r", encoding="utf-8") as file:
        module_ast = parse(file.read())

    for node in module_ast.body:
        if isinstance(node, ClassDef):
            yield node


def exclude_abstract_classes(class_nodes: Iterable[ClassDef]) -> Iterable[ClassDef]:
    abstract_bases = {"Protocol", "Generic", "ABC"}

    def __is_abstract_class_node(node: ClassDef) -> bool:
        for base in node.bases:
            if isinstance(base, Name) and base.id in abstract_bases:
                return True

            if isinstance(base, Subscript) and isinstance(base.value, Name) and base.value.id in abstract_bases:
                return True
        return False

    for class_node in class_nodes:
        if not __is_abstract_class_node(class_node):
            yield class_node


def transform_class_nodes_to_class_data(class_nodes: Iterable[ClassDef], module_file_path: str) -> Iterable[ClassData]:
    for class_node in class_nodes:
        yield ClassData(
            module_file_path=module_file_path,
            class_name=class_node.name,
        )


def exclude_classes_without_public_methods(class_nodes: Iterable[ClassDef]) -> Iterable[ClassDef]:
    for class_node in class_nodes:
        if class_defines_public_methods(class_node):
            yield class_node


def class_defines_public_methods(class_node: ClassDef) -> bool:
    for node in class_node.body:
        if not isinstance(node, FunctionDef) or node.name.startswith("_"):
            continue
        return True
    return False


def exclude_dataclasses(class_nodes: Iterable[ClassDef]) -> Iterable[ClassDef]:
    for class_node in class_nodes:
        if class_has_any_decorator(class_node, {"dataclass"}):
            continue
        yield class_node


def class_has_any_decorator(class_node: ClassDef, decorator_names: Set[str]) -> bool:
    for decorator_node in class_node.decorator_list:
        decorator_id = None
        if isinstance(decorator_node, Name):
            decorator_id = decorator_node.id
        elif isinstance(decorator_node, Call):
            decorator_id = decorator_node.func.id

        if decorator_id in decorator_names:
            return True
    return False


def exclude_classes_without_decorators(
    class_nodes: Iterable[ClassDef], decorator_names: Set[str]
) -> Iterable[ClassDef]:
    for class_node in class_nodes:
        if class_has_any_decorator(class_node, decorator_names):
            yield class_node



from typing import Any, Dict, Optional

from yandil.configuration.environment import Environment


class ConfigurationContainer:
    def __init__(self):
        self.__configurations_map: Dict[str, Any] = {}

    def __getitem__(self, configuration_key: str) -> Optional[Any]:
        configuration_value = self.__configurations_map.get(configuration_key)
        if configuration_value is None:
            return None

        if isinstance(configuration_value, Environment):
            configuration_value = configuration_value.resolve()

        return configuration_value

    def __setitem__(self, configuration_key: str, configuration_value: Any) -> None:
        self.__configurations_map[configuration_key] = configuration_value


default_configuration_container = ConfigurationContainer()



from typing import Type


class AbstractClassNotAllowedError(Exception):
    def __init__(self, abstract_class: Type):
        self.abstract_class = abstract_class
        super().__init__(f"Abstract class {abstract_class.__name__} not allowed")
